import logging
from urllib.parse import parse_qsl, unquote, urlencode, urlparse
from zoneinfo import ZoneInfo

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from edc_dashboard.utils import get_bootstrap_version
from edc_utils import convert_php_dateformat

from .site_action_items import site_action_items

logger = logging.getLogger(__name__)


class ActionItemHelperError(Exception):
    pass


class ActionItemHelper:

    action_item_reason_template_name = (
        f"edc_action_item/bootstrap{get_bootstrap_version()}/action_item_reason.html"
    )

    def __init__(
        self, action_item=None, href=None, action_name=None, related_action_item=None
    ):
        self._parent_reference_obj = None
        self._parent_reference_url = None
        self._reference_obj = None
        self._reference_url = None
        self._related_reference_obj = None
        self._related_reference_url = None
        self.href = href
        self.last_updated_text = "This action item has not been updated."
        if not action_item and action_name:
            self.action_identifier = None
            self.action_item = None
            self.action_cls = site_action_items.get(action_name)
            self.related_action_item = related_action_item
            if self.action_cls.related_reference_fk_attr and not related_action_item:
                raise ActionItemHelperError(
                    f"Expected related_action_item. Got None. "
                    f"Related field attribute is '"
                    f"{self.action_cls.related_reference_fk_attr}'. "
                    f"See {repr(self)}."
                )
        else:
            self.action_identifier = action_item.action_identifier
            self.action_item = action_item
            self.action_cls = action_item.reference_model_cls.get_action_cls()
            self.related_action_item = self.action_item.related_action_item
            if self.action_item.last_updated:
                # could also use action_item.linked_to_reference?
                date_format = convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                last_updated = self.action_item.last_updated.strftime(date_format)
                user_last_updated = self.action_item.user_last_updated
                self.last_updated_text = (
                    f"Last updated on {last_updated} by {user_last_updated}."
                )

    def __repr__(self):
        return f"{self.__class__.__name__}(action_name={self.action_cls})"

    def get_url(self, model_obj=None, model_cls=None, **kwargs):
        """Returns a relative add URL with querystring that can
        get back to the subject dashboard on save.

        Adds visit fk to the querystring if possible.
        """
        opts = dict(**self.get_query_dict(), **kwargs)
        if self.action_identifier:
            opts.update(action_identifier=self.action_identifier)

        if self.action_cls.related_reference_fk_attr:
            try:
                related_reference_obj = (
                    self.action_cls.related_reference_model_cls().objects.get(
                        action_item=self.related_action_item
                    )
                )
            except ObjectDoesNotExist as e:
                logger.warning(
                    f"{e} See {self.action_item}. Related action identifier"
                    f"='{self.action_item.related_action_identifier}'."
                )
                opts.update({self.action_cls.related_reference_fk_attr: None})
            else:
                opts.update(
                    {self.action_cls.related_reference_fk_attr: str(related_reference_obj.pk)}
                )
        if model_obj:
            try:
                model_obj.related_visit_model_attr()
            except AttributeError:
                pass
            else:
                visit_obj = getattr(model_obj, model_obj.related_visit_model_attr())
                opts.update(
                    {
                        model_obj.related_visit_model_attr(): str(visit_obj.pk),
                        "appointment": str(visit_obj.appointment.pk),
                    }
                )
            path = model_obj.get_absolute_url()
        else:
            path = model_cls().get_absolute_url()
        query = unquote(urlencode(opts))
        if query:
            return "?".join([path, query])
        return path

    @property
    def reference_obj(self):
        """Returns the reference model instance or None."""
        if not self._reference_obj:
            try:
                self._reference_obj = self.action_item.reference_obj
            except (AttributeError, ObjectDoesNotExist):
                pass
        return self._reference_obj

    @property
    def reference_url(self):
        """Returns the url for the reference object."""
        if not self._reference_url:
            self._reference_url = self.get_url(
                model_obj=self.reference_obj,
                model_cls=self.action_cls.reference_model_cls(),
            )
        return self._reference_url

    @property
    def parent_reference_obj(self):
        """Returns the parent reference model instance or None."""
        if not self._parent_reference_obj:
            try:
                self._parent_reference_obj = self.action_item.parent_reference_obj
            except (AttributeError, ObjectDoesNotExist):
                pass
        return self._parent_reference_obj

    def get_query_dict(self):
        return dict(parse_qsl(urlparse(self.href).query))

    @property
    def parent_reference_url(self):
        """Returns the change url for the parent reference
        model instance or None.
        """
        if not self._parent_reference_url:
            if self.parent_reference_obj:
                kwargs = dict(
                    model_obj=self.parent_reference_obj,
                    model_cls=self.parent_reference_obj.__class__,
                )
                self._parent_reference_url = self.get_url(**kwargs)
        return self._parent_reference_url

    @property
    def related_reference_obj(self):
        """Returns the change url for the related reference
        model instance or None.
        """
        if not self._related_reference_obj:
            try:
                self._related_reference_obj = self.related_action_item.reference_obj
            except (AttributeError, ObjectDoesNotExist):
                pass
        return self._related_reference_obj

    @property
    def related_reference_url(self):
        """Returns the change url for the related reference
        model instance or None.
        """
        if not self._related_reference_url:
            if self.related_reference_obj:
                kwargs = dict(
                    model_obj=self.related_reference_obj,
                    model_cls=self.related_reference_obj.__class__,
                )
                self._related_reference_url = self.get_url(**kwargs)
        return self._related_reference_url

    def render_action_item_reasons(self):
        action_item_reasons = []
        objects = [
            self.reference_obj,
            self.parent_reference_obj,
            self.related_reference_obj,
        ]
        for obj in objects:
            try:
                action_item_reasons.append(obj.get_action_item_reason())
            except AttributeError as e:
                if "get_action_item_reason" not in str(e):
                    raise
        action_item_reasons = list(set(action_item_reasons))
        return render_to_string(
            self.action_item_reason_template_name,
            context={"action_item_reasons": action_item_reasons},
        )

    @property
    def reference_model_name(self):
        try:
            reference_model_name = (
                f"{self.action_item.reference_model_cls._meta.verbose_name} "
                f'{str(self.reference_obj or "")}'
            )
        except AttributeError:
            reference_model_name = None
        return reference_model_name

    @property
    def parent_reference_model_name(self):
        try:
            parent_reference_model_name = (
                f"{self.action_item.parent_action_item.reference_model_cls._meta.verbose_name} "  # noqa
                f"{str(self.parent_reference_obj)}"
            )
        except AttributeError:
            parent_reference_model_name = None
        return parent_reference_model_name

    @property
    def related_reference_model_name(self):
        try:
            related_reference_model_name = (
                f"{self.related_action_item.reference_model_cls._meta.verbose_name} "
                f"{str(self.related_reference_obj)}"
            )
        except AttributeError:
            related_reference_model_name = None
        return related_reference_model_name

    def get_context(self):
        """Returns a dictionary of instance attr."""
        context = {}
        if self.action_item.parent_action_item:
            context.update(
                parent_action_identifier=self.action_item.parent_action_item.action_identifier
            )
        if self.related_action_item:
            context.update(
                related_action_identifier=self.related_action_item.action_identifier
            )
        context.update(
            action_identifier=self.action_identifier,
            action_instructions=self.action_item.instructions,
            action_item_color=self.action_item.action.get_color_style(),
            action_item_reason=self.render_action_item_reasons(),
            display_name=self.action_item.action.get_display_name(),
            href=self.href,
            last_updated_text=self.last_updated_text,
            name=self.action_item.action_type.name,
            parent_action_item=self.action_item.parent_action_item,
            parent_reference_obj=self.parent_reference_obj,
            parent_reference_model_name=self.parent_reference_model_name,
            parent_reference_url=self.parent_reference_url,
            popover_title=self.action_item.action.get_popover_title(),
            priority=self.action_item.action.get_priority(),
            reference_model_name=self.reference_model_name,
            reference_obj=self.reference_obj,
            reference_url=self.reference_url,
            related_reference_obj=self.related_reference_obj,
            related_reference_model_name=self.related_reference_model_name,
            related_reference_url=self.related_reference_url,
            report_datetime=self.action_item.report_datetime.astimezone(
                ZoneInfo(settings.TIME_ZONE)
            ),
            status=self.action_item.action.get_status(),
        )
        return context



from django import forms

from ..action import ActionError


class ActionItemModelFormMixin:
    def clean(self):
        cleaned_data = super().clean()
        action_cls = self._meta.model.get_action_cls()
        if self.action_cls:
            try:
                action_cls(
                    subject_identifier=cleaned_data.get("subject_identifier"),
                    action_identifier=cleaned_data.get("action_identifier"),
                    related_action_item=cleaned_data.get("related_action_item"),
                    readonly=True,
                )
            except ActionError as e:
                raise forms.ValidationError(
                    f"{str(e.message)}. Please contact your data manager."
                )
        return cleaned_data

    @property
    def action_cls(self):
        try:
            action_cls = self._meta.model.get_action_cls()
        except AttributeError:
            action_cls = None
        return action_cls



# Generated by Django 2.1 on 2018-10-09 02:45

import django.db.models.deletion
from django.db import migrations, models

import edc_action_item.models.action_item


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0009_auto_20180927_0306")]

    operations = [
        migrations.AlterModelManagers(
            name="actionitem",
            managers=[
                ("on_site", edc_action_item.models.action_item.CurrentSiteManager()),
                ("objects", edc_action_item.models.action_item.ActionItemManager()),
            ],
        ),
        migrations.RemoveField(model_name="actionitem", name="parent_reference_model"),
        migrations.RemoveField(
            model_name="historicalactionitem", name="parent_reference_model"
        ),
        migrations.AddField(
            model_name="actionitem",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="reference",
            name="action_item",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="reference",
            name="parent_action_item",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AddField(
            model_name="reference",
            name="related_action_item",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="action_identifier",
            field=models.CharField(max_length=50, unique=True),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="action_identifier",
            field=models.CharField(db_index=True, max_length=50),
        ),
    ]



from django.test import TestCase
from edc_test_utils.natural_key_test_helper import NaturalKeyTestHelper

from edc_action_item.site_action_items import site_action_items

from ..test_case_mixin import TestCaseMixin


class TestNaturalKey(TestCaseMixin, TestCase):

    natural_key_helper = NaturalKeyTestHelper()

    def setUp(self):
        self.subject_identifier = self.fake_enroll()
        site_action_items.registry = {}

    def test_natural_key_attrs(self):
        self.natural_key_helper.nk_test_natural_key_attr(
            "edc_action_item", exclude_models=["edc_action_item.subjectidentifiermodel"]
        )

    def test_get_by_natural_key_attr(self):
        self.natural_key_helper.nk_test_get_by_natural_key_attr(
            "edc_action_item", exclude_models=["edc_action_item.subjectidentifiermodel"]
        )

    # def test_deserialize_action_item(self):
    #     site_action_items.register(FormOneAction)
    #     get_action_type(FormOneAction)
    #     action = FormOneAction(subject_identifier=self.subject_identifier)
    #     action_item = ActionItem.objects.get(action_identifier=action.action_identifier)
    #     for outgoing_transaction in OutgoingTransaction.objects.filter(
    #         tx_name=action_item._meta.label_lower
    #     ):
    #         self.offline_test_helper.offline_test_deserialize(
    #             action_item, outgoing_transaction
    #         )



from django.core.exceptions import ObjectDoesNotExist
from django.db.models.deletion import ProtectedError
from django.test import TestCase
from edc_constants.constants import CLOSED, NEW, OPEN

from edc_action_item.delete_action_item import ActionItemDeleteError, delete_action_item
from edc_action_item.models import ActionItem
from edc_action_item.site_action_items import site_action_items
from edc_action_item.tests.models import FormTwo

from ..action_items import (
    FormOneAction,
    FormThreeAction,
    FormTwoAction,
    SingletonAction,
    register_actions,
)
from ..models import FormOne
from ..test_case_mixin import TestCaseMixin


class TestAction(TestCaseMixin, TestCase):
    def setUp(self):
        register_actions()
        self.subject_identifier = self.fake_enroll()
        self.assertIn(FormOneAction.name, site_action_items.registry)
        self.assertIn(FormTwoAction.name, site_action_items.registry)
        self.assertIn(FormThreeAction.name, site_action_items.registry)

    def test_reference_model_delete_resets_action_item(self):
        obj = FormOne.objects.create(subject_identifier=self.subject_identifier)
        action_item = ActionItem.objects.get(action_identifier=obj.action_identifier)
        self.assertEqual(action_item.status, CLOSED)
        obj.delete()
        action_item = ActionItem.objects.get(action_identifier=obj.action_identifier)
        self.assertEqual(action_item.status, NEW)
        self.assertFalse(action_item.linked_to_reference)

    def test_delete_singleton(self):
        SingletonAction(subject_identifier=self.subject_identifier)
        delete_action_item(
            action_cls=SingletonAction, subject_identifier=self.subject_identifier
        )
        self.assertRaises(
            ObjectDoesNotExist,
            ActionItem.objects.get,
            subject_identifier=self.subject_identifier,
        )

    def test_cannot_delete_if_not_new(self):
        action = SingletonAction(subject_identifier=self.subject_identifier)

        action_item = action.action_item

        action_item.status = CLOSED
        action_item.save()
        self.assertRaises(
            ActionItemDeleteError,
            delete_action_item,
            action_cls=SingletonAction,
            subject_identifier=self.subject_identifier,
        )
        action_item.status = OPEN
        action_item.save()
        self.assertRaises(
            ActionItemDeleteError,
            delete_action_item,
            action_cls=SingletonAction,
            subject_identifier=self.subject_identifier,
        )

    def test_deleted_next_action_item(self):
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        ActionItem.objects.get(action_identifier=form_one.action_identifier)
        form_two = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )

        action_item = ActionItem.objects.get(action_identifier=form_two.action_identifier)
        self.assertRaises(ProtectedError, action_item.delete)



from django.test import TestCase
from edc_facility.import_holidays import import_holidays
from edc_metadata.tests.models import SubjectConsent
from edc_reference import ReferenceModelConfig, site_reference_configs
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules

from .visit_schedule import visit_schedule


class TestCaseMixin(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    @staticmethod
    def enroll(subject_identifier=None):
        subject_identifier = subject_identifier or "1111111"

        site_reference_configs.registry = {}
        reference = ReferenceModelConfig(
            name="edc_action_item.CrfLongitudinalOne", fields=["f1", "f2", "f3"]
        )
        site_reference_configs.register(reference)

        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_metadata.subjectvisit"}
        )

        subject_consent = SubjectConsent.objects.create(
            subject_identifier=subject_identifier, consent_datetime=get_utcnow()
        )
        _, schedule = site_visit_schedules.get_by_onschedule_model("edc_metadata.onschedule")
        schedule.put_on_schedule(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        return subject_identifier

    @staticmethod
    def fake_enroll():
        subject_identifier = "2222222"
        RegisteredSubject.objects.create(subject_identifier=subject_identifier)
        return subject_identifier



# Generated by Django 2.2.2 on 2019-06-28 19:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0019_auto_20190627_2257")]

    operations = [
        migrations.AlterModelOptions(name="actiontype", options={"ordering": ["name"]}),
        migrations.AddIndex(
            model_name="actiontype",
            index=models.Index(fields=["name"], name="edc_action__name_f674a7_idx"),
        ),
    ]



from django.apps import apps as django_apps
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_visit_tracking.constants import SCHEDULED

from edc_action_item.helpers import ActionItemHelper
from edc_action_item.model_wrappers import ActionItemModelWrapper
from edc_action_item.models import ActionItem

from ..action_items import CrfOneAction, FormOneAction, register_actions
from ..models import CrfOne, CrfTwo, FormOne, FormTwo
from ..test_case_mixin import TestCaseMixin


class TestHelpers(TestCaseMixin, TestCase):
    def setUp(self):
        register_actions()
        self.subject_identifier = self.fake_enroll()
        self.form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        self.action_item = ActionItem.objects.get(
            action_identifier=self.form_one.action_identifier
        )
        self.model_wrapper = ActionItemModelWrapper(model_obj=self.action_item)

    def test_init(self):
        helper = ActionItemHelper(
            action_item=self.model_wrapper.object, href=self.model_wrapper.href
        )
        self.assertTrue(helper.get_context())

    def test_new_action(self):

        action = CrfOneAction(subject_identifier=self.subject_identifier)

        self.assertIsNone(action.reference_obj)

        model_wrapper = ActionItemModelWrapper(model_obj=action.action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertIsNone(context["reference_obj"])
        self.assertIsNone(context["parent_reference_obj"])
        self.assertIsNone(context["related_reference_obj"])
        self.assertTrue(context["reference_url"])
        self.assertIsNone(context["parent_reference_url"])
        self.assertIsNone(context["related_reference_url"])

    def test_new_action_url(self):
        action = FormOneAction(subject_identifier=self.subject_identifier)
        action_item = action.action_item
        model_wrapper = ActionItemModelWrapper(model_obj=action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertEqual(
            context["reference_url"].split("?")[0],
            "/admin/edc_action_item/formone/add/",
        )
        self.assertEqual(
            context["reference_url"].split("?")[1],
            "next=edc_action_item:subject_dashboard_url,"
            f"subject_identifier&subject_identifier={self.subject_identifier}&"
            f"action_identifier={action_item.action_identifier}",
        )

    def test_action_with_reference_model_instance(self):
        helper = ActionItemHelper(
            action_item=self.model_wrapper.object, href=self.model_wrapper.href
        )
        context = helper.get_context()
        self.assertTrue(context["reference_obj"])
        self.assertIsNone(context["parent_reference_obj"])
        self.assertIsNone(context["related_reference_obj"])
        self.assertTrue(context["reference_url"])
        self.assertIsNone(context["parent_reference_url"])
        self.assertIsNone(context["related_reference_url"])

    def test_reference_obj_and_url(self):

        helper = ActionItemHelper(
            action_item=self.model_wrapper.object, href=self.model_wrapper.href
        )
        context = helper.get_context()
        self.assertEqual(context["reference_obj"], self.form_one)
        self.assertEqual(
            context["reference_url"].split("?")[0],
            f"/admin/edc_action_item/formone/{str(self.form_one.id)}/change/",
        )

        self.assertEqual(
            context["reference_url"].split("?")[1],
            "next=edc_action_item:subject_dashboard_url,"
            f"subject_identifier&subject_identifier={self.subject_identifier}&"
            f"action_identifier={self.action_item.action_identifier}",
        )

    def test_create_parent_reference_model_instance_then_delete(self):
        form_two = FormTwo.objects.create(
            form_one=self.form_one, subject_identifier=self.subject_identifier
        )
        action_item = ActionItem.objects.get(action_identifier=form_two.action_identifier)

        model_wrapper = ActionItemModelWrapper(model_obj=action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertEqual(context["reference_obj"], form_two)
        form_two.delete()
        action_item = ActionItem.objects.get(action_identifier=form_two.action_identifier)
        model_wrapper = ActionItemModelWrapper(model_obj=action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertIsNone(context["reference_obj"])

    def test_create_parent_reference_model_instance(self):
        form_two = FormTwo.objects.create(
            form_one=self.form_one, subject_identifier=self.subject_identifier
        )
        action_item = ActionItem.objects.get(action_identifier=form_two.action_identifier)
        model_wrapper = ActionItemModelWrapper(model_obj=action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertEqual(context["reference_obj"], form_two)
        self.assertEqual(context["parent_reference_obj"], self.form_one)
        self.assertEqual(context["related_reference_obj"], self.form_one)
        self.assertTrue(context["reference_url"])
        self.assertTrue(context["parent_reference_url"])
        self.assertTrue(context["related_reference_url"])

    def test_create_next_parent_reference_model_instance(self):
        first_form_two = FormTwo.objects.create(
            form_one=self.form_one, subject_identifier=self.subject_identifier
        )
        second_form_two = FormTwo.objects.create(
            form_one=self.form_one, subject_identifier=self.subject_identifier
        )
        action_item = ActionItem.objects.get(
            action_identifier=second_form_two.action_identifier
        )
        model_wrapper = ActionItemModelWrapper(model_obj=action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertEqual(context["reference_obj"], second_form_two)
        self.assertEqual(context["parent_reference_obj"], first_form_two)
        self.assertEqual(context["related_reference_obj"], self.form_one)
        self.assertTrue(context["reference_url"])
        self.assertTrue(context["parent_reference_url"])
        self.assertTrue(context["related_reference_url"])

    def test_reference_as_crf(self):
        self.enroll()
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = django_apps.get_model("edc_metadata.subjectvisit").objects.create(
            appointment=appointment, reason=SCHEDULED
        )
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        action_item = ActionItem.objects.get(action_identifier=crf_one.action_identifier)
        model_wrapper = ActionItemModelWrapper(model_obj=action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertEqual(context["reference_obj"], crf_one)
        self.assertIsNone(context["parent_reference_obj"])
        self.assertIsNone(context["related_reference_obj"])
        self.assertTrue(context["reference_url"])
        self.assertIsNone(context["parent_reference_url"])
        self.assertIsNone(context["related_reference_url"])

    def test_reference_as_crf_create_next_model_instance(self):
        subject_identifier = self.enroll()
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = django_apps.get_model("edc_metadata.subjectvisit").objects.create(
            appointment=appointment, reason=SCHEDULED
        )
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        crf_two = CrfTwo.objects.create(subject_visit=subject_visit)
        action_item = ActionItem.objects.get(action_identifier=crf_two.action_identifier)
        model_wrapper = ActionItemModelWrapper(model_obj=action_item)
        helper = ActionItemHelper(action_item=model_wrapper.object, href=model_wrapper.href)
        context = helper.get_context()
        self.assertEqual(context["reference_obj"], crf_two)
        self.assertEqual(context["parent_reference_obj"], crf_one)
        self.assertIsNone(context["related_reference_obj"])
        self.assertTrue(context["reference_url"])
        self.assertTrue(context["parent_reference_url"])
        self.assertIsNone(context["related_reference_url"])

        self.assertEqual(
            context["parent_reference_url"].split("?")[0],
            f"/admin/edc_action_item/crfone/{str(crf_one.id)}/change/",
        )
        self.assertEqual(
            context["parent_reference_url"].split("?")[1],
            "next=edc_action_item:subject_dashboard_url,"
            f"subject_identifier&subject_identifier={subject_identifier}&"
            f"action_identifier={action_item.action_identifier}&"
            f"subject_visit={str(subject_visit.pk)}&"
            f"appointment={str(appointment.pk)}",
        )



from django.core.exceptions import ObjectDoesNotExist

from .get_action_type import get_action_type


class SingletonActionItemError(Exception):
    pass


class CreateActionItemError(Exception):
    pass


def create_action_item(
    action_cls,
    subject_identifier=None,
    action_identifier=None,
    related_action_item=None,
    parent_action_item=None,
    priority=None,
    using=None,
    **kwargs,  # noqa
):
    action_item = None
    if action_cls.singleton:
        try:
            action_item = (
                action_cls.action_item_model_cls()
                .objects.using(using)
                .get(
                    action_type=get_action_type(action_cls),
                    subject_identifier=subject_identifier,
                )
            )
        except ObjectDoesNotExist:
            pass
        else:
            raise SingletonActionItemError(
                f"Action {action_cls.name} can only be "
                f"created once per subject. Got {subject_identifier}."
            )
    if not action_item:
        opts = dict(
            subject_identifier=subject_identifier,
            action_type=get_action_type(action_cls),
            action_identifier=action_identifier,
            linked_to_reference=False,
            priority=priority,
        )
        if parent_action_item:
            opts.update(parent_action_item=parent_action_item)
        if related_action_item:
            opts.update(related_action_item=related_action_item)
        action_item = action_cls.action_item_model_cls()(**opts)
        action_item.save(using=using)
        action_item.refresh_from_db()
    return action_item



from django.contrib import admin
from django.urls.base import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django_audit_fields.admin import audit_fieldset_tuple
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_model_admin.history import SimpleHistoryAdmin

from ..admin_site import edc_action_item_admin
from ..forms import ActionItemForm
from ..models import ActionItem


@admin.register(ActionItem, site=edc_action_item_admin)
class ActionItemAdmin(ModelAdminSubjectDashboardMixin, SimpleHistoryAdmin):

    form = ActionItemForm

    save_on_top = True
    show_cancel = True

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "action_identifier",
                    "subject_identifier",
                    "report_datetime",
                    "action_type",
                    "priority",
                    "status",
                    "instructions",
                )
            },
        ),
        (
            "Reference Information",
            {
                "classes": ("collapse",),
                "fields": (
                    "related_action_item",
                    "parent_action_item",
                    "auto_created",
                    "auto_created_comment",
                ),
            },
        ),
        (
            "Email",
            {"classes": ("collapse",), "fields": ("emailed", "emailed_datetime")},
        ),
        audit_fieldset_tuple,
    )

    radio_fields = {"status": admin.VERTICAL}

    list_display = (
        "identifier",
        "dashboard",
        "subject_identifier",
        "status",
        "action_type",
        "priority",
        "emailed",
        "parent_action",
        "related_action_item",
        "created",
    )

    list_filter = (
        "status",
        "priority",
        "emailed",
        "report_datetime",
        "action_type__name",
    )

    search_fields = (
        "subject_identifier",
        "action_identifier",
        "related_action_item__action_identifier",
        "parent_action_item__action_identifier",
        "action_type__name",
        "action_type__display_name",
        "id",
    )

    ordering = ("action_type__display_name",)

    date_hierarchy = "created"

    additional_instructions = format_html(
        "<B><U>Important:</U> This form is usually auto-filled based on a clinical event. "
        "DO NOT DELETE unless you know what you are doing.</B>"
    )

    def get_readonly_fields(self, request, obj=None) -> tuple:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        readonly_fields = readonly_fields + (
            "action_identifier",
            "instructions",
            "auto_created",
            "auto_created_comment",
            "reference_model",
            "emailed",
            "emailed_datetime",
            "related_action_item",
            "parent_action_item",
        )
        if obj:
            readonly_fields = readonly_fields + (
                "subject_identifier",
                "report_datetime",
                "action_type",
            )
        return readonly_fields

    @staticmethod
    def parent_action(obj):
        """Returns an url to the parent action item
        for display in admin.
        """
        if obj.parent_action_item:
            url_name = "_".join(obj._meta.label_lower.split("."))
            namespace = edc_action_item_admin.name
            url = reverse(f"{namespace}:{url_name}_changelist")
            return format_html(
                '<a data-toggle="tooltip" title="go to parent action item" '
                'href="{}?q={}">{}</a>',
                mark_safe(url),  # nosec B308, B703
                mark_safe(obj.parent_action_item.action_identifier),  # nosec B308, B703
                mark_safe(obj.parent_action_item.identifier),  # nosec B308, B703
            )
        return None

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "action_type":
            kwargs["queryset"] = db_field.related_model.objects.filter(create_by_user=True)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)



# Generated by Django 2.0.1 on 2018-01-09 09:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0001_initial")]

    operations = [
        migrations.AlterModelOptions(
            name="actionitem",
            options={
                "verbose_name": "Action Item",
                "verbose_name_plural": "Action Items",
            },
        )
    ]



# Generated by Django 2.2.2 on 2019-06-28 19:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0021_auto_20190628_2113")]

    operations = [
        migrations.RemoveIndex(model_name="actionitem", name="edc_action__action__fe58cb_idx"),
        migrations.RemoveIndex(model_name="actiontype", name="edc_action__name_f674a7_idx"),
        migrations.AddIndex(
            model_name="actionitem",
            index=models.Index(
                fields=[
                    "id",
                    "action_identifier",
                    "action_type",
                    "status",
                    "report_datetime",
                ],
                name="edc_action__id_6cb4e0_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="actiontype",
            index=models.Index(fields=["id", "name"], name="edc_action__id_4feab7_idx"),
        ),
    ]



from edc_visit_schedule import site_visit_schedules
from edc_visit_schedule.models import SubjectScheduleHistory


class VisitScheduleActionMixin:

    """A mixin for the Action class to add properties and
    methods related to the visit schedule from
    edc_visit_schedule.
    """

    def __init__(self, reference_obj=None, subject_identifier=None, **kwargs):
        super().__init__(
            reference_obj=reference_obj, subject_identifier=subject_identifier, **kwargs
        )
        self.offschedule_model = None
        if reference_obj and subject_identifier:
            for onschedule_model_obj in SubjectScheduleHistory.objects.onschedules(
                subject_identifier=subject_identifier,
                report_datetime=reference_obj.report_datetime,
            ):
                _, schedule = site_visit_schedules.get_by_onschedule_model(
                    onschedule_model=onschedule_model_obj._meta.label_lower
                )
                self.offschedule_model = schedule.offschedule_model



# Generated by Django 2.1.3 on 2018-11-21 15:38

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0013_auto_20181108_0353")]

    operations = [
        migrations.RemoveField(model_name="actionitemupdate", name="action_item"),
        migrations.RemoveField(model_name="historicalactionitemupdate", name="action_item"),
        migrations.RemoveField(model_name="historicalactionitemupdate", name="history_user"),
        migrations.DeleteModel(name="ActionItemUpdate"),
        migrations.DeleteModel(name="HistoricalActionItemUpdate"),
    ]



# Generated by Django 2.2.2 on 2019-06-28 19:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0020_auto_20190628_2113")]

    operations = [
        migrations.RemoveIndex(model_name="actionitem", name="edc_action__action__b19a08_idx"),
        migrations.AddIndex(
            model_name="actionitem",
            index=models.Index(
                fields=[
                    "action_identifier",
                    "action_type",
                    "status",
                    "report_datetime",
                ],
                name="edc_action__action__fe58cb_idx",
            ),
        ),
    ]



# Generated by Django 2.0.2 on 2018-03-19 21:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0003_auto_20180116_1528")]

    operations = [
        migrations.RenameField(
            model_name="actionitem",
            old_name="parent_model",
            new_name="parent_reference_model",
        ),
        migrations.RenameField(
            model_name="historicalactionitem",
            old_name="parent_model",
            new_name="parent_reference_model",
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="model",
            field=models.CharField(
                blank=True, help_text="reference model", max_length=100, null=True
            ),
        ),
    ]



from edc_constants.constants import HIGH_PRIORITY, NO

from ..action import Action
from ..action_with_notification import ActionWithNotification
from ..site_action_items import site_action_items


class FormZeroAction(ActionWithNotification):
    name = "submit-form-zero"
    display_name = "Submit Form Zero"
    reference_model = "edc_action_item.formzero"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    parent_action_names = ["submit-form-three"]

    notification_display_name = "a form zero event has occured"
    notification_fields = ["f1"]
    notification_email_to = ["someone@example.com"]


class TestDoNothingPrnAction(Action):

    name = "test-nothing-prn-action"
    display_name = "Test Nothing Prn Action"
    parent_action_names = None


class TestPrnAction(Action):

    name = "test-prn-action"
    display_name = "Test Prn Action"
    next_actions = [FormZeroAction.name]
    parent_action_names = None


class FormThreeAction(Action):
    name = "submit-form-three"
    display_name = "Submit Form Three"
    reference_model = "edc_action_item.formthree"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = [FormZeroAction.name]
    parent_action_names = ["submit-form-one"]


class FormTwoAction(Action):
    name = "submit-form-two"
    display_name = "Submit Form Two"
    reference_model = "edc_action_item.formtwo"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    related_reference_model = "edc_action_item.formone"
    related_reference_fk_attr = "form_one"
    next_actions = ["self"]
    parent_action_names = ["submit-form-two", "submit-form-one"]


class FormOneAction(Action):
    name = "submit-form-one"
    display_name = "Submit Form One"
    reference_model = "edc_action_item.formone"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = [FormTwoAction.name, FormThreeAction.name]
    parent_action_names = ["submit-form-three", "submit-form-four"]


class FormFourAction(Action):
    name = "submit-form-four"
    display_name = "Submit Form Four"
    reference_model = "edc_action_item.formfour"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    parent_action_names = ["submit-form-one"]

    def get_next_actions(self):
        next_actions = []
        next_actions = self.append_to_next_if_required(
            next_actions=next_actions,
            action_name=FormOneAction.name,
            required=self.reference_obj.happy == NO,
        )
        return next_actions


class FollowupAction(Action):
    name = "submit-followup"
    display_name = "Submit Followup"
    reference_model = "edc_action_item.followup"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = ["self"]
    related_reference_fk_attr = "initial"
    related_reference_model = "edc_action_item.initial"
    parent_action_names = ["submit-followup", "submit-initial"]


class CrfTwoAction(Action):
    name = "submit-crf-two"
    display_name = "Submit Crf Two"
    reference_model = "edc_action_item.crftwo"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = ["self"]
    parent_action_names = ["submit-crf-two", "submit-crf-one"]


class CrfOneAction(Action):
    name = "submit-crf-one"
    display_name = "Submit Crf One"
    reference_model = "edc_action_item.crfone"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = [CrfTwoAction.name]


class InitialAction(Action):
    name = "submit-initial"
    display_name = "Submit Initial"
    reference_model = "edc_action_item.initial"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = [FollowupAction.name]


class SingletonAction(Action):
    name = "singleton"
    display_name = "Singleton"
    reference_model = "edc_action_item.formzero"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    singleton = True


class CrfLongitudinalTwoAction(Action):
    name = "submit-crf-longitudinal-two"
    display_name = "Submit Crf Two"
    reference_model = "edc_action_item.crflongitudinaltwo"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = None
    parent_action_names = ["submit-crf-longitudinal-one"]


class CrfLongitudinalOneAction(Action):
    name = "submit-crf-longitudinal-one"
    display_name = "Submit Crf One"
    reference_model = "edc_action_item.crflongitudinalone"
    show_on_dashboard = True
    priority = HIGH_PRIORITY
    next_actions = [CrfLongitudinalTwoAction.name]


def register_actions():
    site_action_items.registry = {}
    site_action_items.register(FormZeroAction)
    site_action_items.register(FormOneAction)
    site_action_items.register(FormTwoAction)
    site_action_items.register(FormThreeAction)
    site_action_items.register(FormFourAction)
    site_action_items.register(InitialAction)
    site_action_items.register(FollowupAction)
    site_action_items.register(TestDoNothingPrnAction)
    site_action_items.register(TestPrnAction)
    site_action_items.register(SingletonAction)
    site_action_items.register(CrfOneAction)
    site_action_items.register(CrfTwoAction)
    site_action_items.register(CrfLongitudinalOneAction)
    site_action_items.register(CrfLongitudinalTwoAction)



from typing import Tuple

from .fieldsets import action_fields


class ActionItemModelAdminMixin:
    def get_readonly_fields(self, request, obj=None) -> Tuple[str, ...]:
        """
        Returns a list of readonly field names.

        Note: "action_identifier" is remove.
            You are expected to use ActionItemFormMixin with the form.
        """
        fields = super().get_readonly_fields(request, obj=obj)
        fields += action_fields
        return tuple(f for f in fields if f != "action_identifier")

    def get_search_fields(self, request) -> Tuple[str, ...]:
        search_fields = super().get_search_fields(request)
        custom_fields = ("action_identifier",)
        if "subject_identifier" in self.model._meta.fields:
            custom_fields = ("subject_identifier",) + custom_fields
        return tuple(set(custom_fields + search_fields))



from edc_model_wrapper import ModelWrapper

from ..helpers import ActionItemHelper
from ..models import ActionItem


class ActionItemModelWrapper(ModelWrapper):

    model_cls = ActionItem
    next_url_attrs = ["subject_identifier"]
    next_url_name = "subject_dashboard_url"

    def __init__(self, model_obj=None, **kwargs):
        super().__init__(model_obj=model_obj, **kwargs)
        helper = ActionItemHelper(action_item=self.object, href=self.href)
        for key, value in helper.get_context().items():
            try:
                setattr(self, key, value)
            except AttributeError as e:
                raise AttributeError(f"{e}. Got {key}, {value}")

    @property
    def subject_identifier(self):
        return self.object.subject_identifier

    @property
    def report_date(self):
        return self.object.report_datetime.date()

    @property
    def str_pk(self):
        return str(self.object.id)



from edc_auth.auth_objects import (
    AUDITOR_ROLE,
    CLINICIAN_ROLE,
    CLINICIAN_SUPER_ROLE,
    NURSE_ROLE,
)
from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

from .auth_objects import (
    ACTION_ITEM,
    ACTION_ITEM_EXPORT,
    ACTION_ITEM_VIEW_ONLY,
    action_items_codenames,
    navbar_tuples,
)

site_auths.add_post_update_func(
    "edc_action_item",
    remove_default_model_permissions_from_edc_permissions,
)

site_auths.add_custom_permissions_tuples(
    model="edc_action_item.edcpermissions", codename_tuples=navbar_tuples
)

site_auths.add_group(*action_items_codenames, name=ACTION_ITEM)
site_auths.add_group(*action_items_codenames, name=ACTION_ITEM_VIEW_ONLY, view_only=True)
site_auths.add_group(
    "edc_action_item.export_actionitem",
    "edc_action_item.export_actiontype",
    "edc_action_item.export_historicalactionitem",
    name=ACTION_ITEM_EXPORT,
)

site_auths.add_custom_permissions_tuples(
    model="edc_action_item.historicalactionitem",
    codename_tuples=[
        (
            "edc_action_item.export_historicalactionitem",
            "Cane export historicalactionitem",
        ),
        (
            "edc_action_item.export_historicalreference",
            "Cane export historicalreference",
        ),
    ],
)

site_auths.update_role(ACTION_ITEM, name=CLINICIAN_ROLE)
site_auths.update_role(ACTION_ITEM, name=NURSE_ROLE)
site_auths.update_role(ACTION_ITEM, name=CLINICIAN_SUPER_ROLE)
site_auths.update_role(ACTION_ITEM_VIEW_ONLY, name=AUDITOR_ROLE)



from importlib import import_module

from django.test import TestCase, override_settings
from edc_adverse_event.auth_objects import TMG_ROLE
from edc_auth.auth_updater import AuthUpdater
from edc_auth.site_auths import site_auths
from edc_data_manager.auth_objects import DATA_MANAGER_ROLE, SITE_DATA_MANAGER_ROLE
from edc_export.auth_objects import EXPORT


class TestAuths(TestCase):
    @override_settings(
        EDC_AUTH_SKIP_SITE_AUTHS=True,
        EDC_AUTH_SKIP_AUTH_UPDATER=False,
    )
    def test_load(self):
        site_auths.initialize()
        AuthUpdater.add_empty_groups_for_tests(EXPORT)
        AuthUpdater.add_empty_roles_for_tests(
            TMG_ROLE, DATA_MANAGER_ROLE, SITE_DATA_MANAGER_ROLE
        )
        import_module("edc_action_item.auths")
        AuthUpdater(verbose=True)



import copy
import sys
from importlib import import_module

from django.apps import apps as django_apps
from django.core.management.color import color_style
from django.utils.module_loading import module_has_submodule
from edc_notification import AlreadyRegistered as NotificationAlreadyRegistered
from edc_notification import site_notifications
from edc_prn.prn import Prn
from edc_prn.site_prn_forms import AlreadyRegistered as PrnAlreadyRegistered
from edc_prn.site_prn_forms import site_prn_forms

from .create_or_update_action_type import create_or_update_action_type
from .get_action_type import get_action_type


class AlreadyRegistered(Exception):
    pass


class SiteActionError(Exception):
    pass


class SiteActionItemCollection:
    def __init__(self):
        self.registry = {}
        self.updated_action_types = None
        prn = Prn(model="edc_action_item.actionitem", url_namespace="edc_action_item_admin")
        site_prn_forms.register(prn)

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __len__(self):
        return len(self.registry.values())

    def __iter__(self):
        return iter(self.registry.values())

    def register(self, action_cls=None):
        if action_cls.name in self.registry:
            raise AlreadyRegistered(
                f"Action class is already registered. Got name='{action_cls.name}' "
                f"for {action_cls.__name__}"
            )
        else:
            self.registry.update({action_cls.name: action_cls})
        if action_cls.show_link_to_changelist:
            prn = Prn(
                model=action_cls.get_reference_model(),
                url_namespace=action_cls.admin_site_name,
            )
            try:
                site_prn_forms.register(prn)
            except PrnAlreadyRegistered:
                pass
        try:
            action_cls.notification_email_to
        except AttributeError:
            pass
        else:
            self.register_notifications(action_cls)

    @staticmethod
    def register_notifications(action_cls):
        """Registers a new model notification and an updated model
        notification for this action cls.

        See edc_notification.
        """
        if action_cls.notification_cls():
            try:
                site_notifications.register(action_cls.notification_cls())
            except NotificationAlreadyRegistered:
                pass

    def get(self, name):
        """Returns an action class."""
        if name not in self.registry:
            raise SiteActionError(
                f"Action does not exist. Did you register the Action? "
                f"Expected one of {self.registry}. Got {name}."
            )
        return self.registry.get(name)

    def get_by_model(self, model=None):
        """Returns the action_cls linked to this reference model."""
        for action_cls in self.registry.values():
            if action_cls.get_reference_model() == model:
                return self.get(action_cls.name)
        return None

    def get_show_link_to_add_actions(self):
        class Wrapper:
            def __init__(self, action_cls=None):
                self.name = action_cls.name
                self.display_name = action_cls.display_name
                self.action_type_id = str(get_action_type(action_cls).pk)

        names = [v.name for v in self.registry.values() if v.show_link_to_add]
        return [Wrapper(action_cls=self.get(name)) for name in names]

    def create_or_update_action_types(self):
        """Populates the ActionType model."""
        for action_cls in self.registry.values():
            create_or_update_action_type(**action_cls.as_dict())
        self.updated_action_types = True

    @staticmethod
    def autodiscover(module_name=None, verbose=True):
        before_import_registry = {}
        module_name = module_name or "action_items"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for site {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}           \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_action_items.registry)
                    import_module(f"{app}.{module_name}")
                    writer(f" * registered '{module_name}' from '{app}'\n")
                except SiteActionError as e:
                    writer(f"   - loading {app}.{module_name} ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_action_items.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SiteActionError(str(e))
            except ImportError:
                pass
            except Exception as e:
                raise SiteActionError(
                    f"{e.__class__.__name__} was raised when loading {module_name}. "
                    f"Got {e} See {app}.{module_name}"
                )


site_action_items = SiteActionItemCollection()



from django.db import models
from edc_sites.models import CurrentSiteManager as BaseCurrentSiteManager


class ActionIdentifierSiteManager(BaseCurrentSiteManager):

    use_in_migrations = True

    def get_by_natural_key(self, action_identifier):
        return self.get(action_identifier=action_identifier)


class ActionIdentifierModelManager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, action_identifier):
        return self.get(action_identifier=action_identifier)



from django import forms
from edc_constants.constants import NEW, OPEN

from ..models import ActionItem


class ActionItemForm(forms.ModelForm):

    subject_identifier = forms.CharField(
        label="Subject Identifier",
        required=False,
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    def clean(self):
        cleaned_data = super().clean()
        self.force_open_status()
        return cleaned_data

    def force_open_status(self):
        """Sets status to open for edited NEW action items."""
        if self.instance.id and self.cleaned_data.get("status") == NEW:
            self.cleaned_data["status"] = OPEN

    class Meta:
        model = ActionItem
        fields = "__all__"



# Generated by Django 2.0.3 on 2018-04-09 10:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0004_auto_20180319_2128")]

    operations = [
        migrations.AddField(
            model_name="actionitem",
            name="related_reference_identifier",
            field=models.CharField(
                blank=True,
                help_text="May be left blank. e.g. tracking identifier from source model that opened the item.",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="actionitem",
            name="related_reference_model",
            field=models.CharField(editable=False, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="actiontype",
            name="related_reference_model",
            field=models.CharField(
                blank=True,
                help_text="origin reference model",
                max_length=100,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="related_reference_identifier",
            field=models.CharField(
                blank=True,
                help_text="May be left blank. e.g. tracking identifier from source model that opened the item.",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="related_reference_model",
            field=models.CharField(editable=False, max_length=100, null=True),
        ),
    ]



# Generated by Django 2.1 on 2018-09-27 01:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0008_auto_20180809_0303")]

    operations = [
        migrations.AddField(
            model_name="actionitem",
            name="emailed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="actionitem",
            name="emailed_datetime",
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="emailed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="emailed_datetime",
            field=models.DateTimeField(null=True),
        ),
    ]



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_model_admin.history import SimpleHistoryAdmin

from ..admin_site import edc_action_item_admin
from ..forms import ActionTypeForm
from ..models import ActionType


@admin.register(ActionType, site=edc_action_item_admin)
class ActionTypeAdmin(ModelAdminSubjectDashboardMixin, SimpleHistoryAdmin):

    form = ActionTypeForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "display_name",
                    "model",
                    "show_on_dashboard",
                    "create_by_action",
                    "create_by_user",
                    "instructions",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "name",
        "display_name",
        "show_on_dashboard",
        "create_by_action",
        "create_by_user",
    )

    list_filter = ("create_by_action", "create_by_user", "show_on_dashboard")

    search_fields = ("name", "display_name", "model")

    date_hierarchy = None

    def get_readonly_fields(self, request, obj=None) -> tuple:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return readonly_fields + (
            "name",
            "display_name",
            "model",
            "show_on_dashboard",
            "create_by_action",
            "create_by_user",
            "instructions",
        )



# Generated by Django 2.1 on 2018-10-09 02:45

import django.db.models.deletion
from django.db import migrations, models

import edc_action_item.models.action_item


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0010_auto_20181009_0445")]

    operations = [
        migrations.RenameField(
            model_name="actionitem",
            old_name="parent_reference_identifier",
            new_name="parent_action_identifier",
        ),
        migrations.RenameField(
            model_name="actionitem",
            old_name="related_reference_identifier",
            new_name="related_action_identifier",
        ),
        migrations.RenameField(
            model_name="historicalactionitem",
            old_name="parent_reference_identifier",
            new_name="parent_action_identifier",
        ),
        migrations.RenameField(
            model_name="historicalactionitem",
            old_name="related_reference_identifier",
            new_name="related_action_identifier",
        ),
        migrations.RenameField(
            model_name="reference",
            old_name="parent_reference_identifier",
            new_name="parent_action_identifier",
        ),
        migrations.RenameField(
            model_name="reference",
            old_name="related_reference_identifier",
            new_name="related_action_identifier",
        ),
    ]



# Generated by Django 2.1 on 2018-10-10 21:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0011_auto_20181009_2236")]

    operations = [
        migrations.AlterField(
            model_name="reference",
            name="parent_action_identifier",
            field=models.CharField(
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="related_action_identifier",
            field=models.CharField(
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
    ]



from typing import Optional, Protocol, Type
from uuid import UUID

from django.db import models


class ActionStub(Protocol):
    def __init__(
        self,
        action_item: Optional["ActionItemStub"] = None,
        reference_obj: Optional[models.Model] = None,
        subject_identifier: str = None,
        action_identifier: str = None,
        parent_action_item: Optional["ActionItemStub"] = None,
        related_action_item: Optional["ActionItemStub"] = None,
        using: Optional[str] = None,
        readonly: Optional[bool] = None,
    ) -> None:
        ...

    name: str
    action_item: "ActionItemStub"
    action_identifier: str
    related_reference_model: Optional[str]
    related_reference_fk_attr: Optional[str]
    subject_identifier: str
    related_action_item: Optional["ActionItemStub"]
    ...


class ActionTypeStub(Protocol):
    name: str
    display_name: str
    reference_model: Optional[str]
    related_reference_model: Optional[str]
    ...


class ActionItemStub(Protocol):
    id: UUID
    action_name: str
    action_item_model: str
    subject_dashboard_url: str
    action_identifier: str
    action_item: "ActionItemStub"
    action_type: ActionTypeStub
    linked_to_reference: bool
    parent_action_item: Optional["ActionItemStub"]
    related_action_item: Optional["ActionItemStub"]
    subject_identifier: str
    action_item_reason: str

    action_cls: ActionStub

    def get_status_display(self) -> str:
        ...

    def get_action_cls(self) -> Type[ActionStub]:
        ...


class ActionItemWithNotificationStub(ActionItemStub, Protocol):
    notification_email_to: str
    notification_display_name: str
    notification_fields: list
    notify_on_changed_reference_obj: bool = True
    notify_on_close: bool = False
    notify_on_new: bool = False
    notify_on_new_and_no_reference_obj: bool = True
    notify_on_open: bool = False
    ...



# Generated by Django 2.0 on 2017-12-21 20:15

import _socket
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import edc_model_fields.fields.hostname_modification_field
import edc_utils
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="ActionItem",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(max_length=50, verbose_name="Subject Identifier"),
                ),
                ("action_identifier", models.CharField(max_length=25, unique=True)),
                (
                    "report_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                (
                    "reference_identifier",
                    models.CharField(
                        help_text="e.g. tracking identifier updated from the reference model",
                        max_length=50,
                        null=True,
                    ),
                ),
                ("reference_model", models.CharField(max_length=50, null=True)),
                (
                    "parent_reference_identifier",
                    models.CharField(
                        blank=True,
                        help_text="May be left blank. e.g. tracking identifier from source model that opened the item.",
                        max_length=50,
                        null=True,
                    ),
                ),
                (
                    "parent_model",
                    models.CharField(editable=False, max_length=100, null=True),
                ),
                (
                    "priority",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("high", "High"),
                            ("medium", "Medium"),
                            ("low", "Low"),
                        ],
                        help_text="Leave blank to use default for this action type.",
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("New", "New"),
                            ("open", "Open"),
                            ("closed", "Closed"),
                            ("cancelled", "Cancelled"),
                        ],
                        default="New",
                        max_length=25,
                    ),
                ),
                (
                    "instructions",
                    models.TextField(
                        blank=True, help_text="populated by action class", null=True
                    ),
                ),
                ("auto_created", models.BooleanField(default=False)),
                (
                    "auto_created_comment",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ActionItemUpdate",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "report_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                ("comment", models.TextField(blank=True, max_length=250, null=True)),
                (
                    "action_item",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_action_item.ActionItem",
                    ),
                ),
            ],
            options={
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ActionType",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=50, unique=True)),
                ("display_name", models.CharField(max_length=100)),
                ("model", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "priority",
                    models.CharField(
                        choices=[
                            ("high", "High"),
                            ("medium", "Medium"),
                            ("low", "Low"),
                        ],
                        default="high",
                        max_length=25,
                    ),
                ),
                ("show_on_dashboard", models.BooleanField(default=False)),
                ("show_link_to_changelist", models.BooleanField(default=False)),
                (
                    "create_by_action",
                    models.BooleanField(
                        default=True,
                        help_text="This action may be created by another action",
                    ),
                ),
                (
                    "create_by_user",
                    models.BooleanField(
                        default=True, help_text="This action may be created by the user"
                    ),
                ),
                (
                    "instructions",
                    models.TextField(blank=True, max_length=250, null=True),
                ),
            ],
            options={
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="HistoricalActionItem",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(max_length=50, verbose_name="Subject Identifier"),
                ),
                ("action_identifier", models.CharField(db_index=True, max_length=25)),
                (
                    "report_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                (
                    "reference_identifier",
                    models.CharField(
                        help_text="e.g. tracking identifier updated from the reference model",
                        max_length=50,
                        null=True,
                    ),
                ),
                ("reference_model", models.CharField(max_length=50, null=True)),
                (
                    "parent_reference_identifier",
                    models.CharField(
                        blank=True,
                        help_text="May be left blank. e.g. tracking identifier from source model that opened the item.",
                        max_length=50,
                        null=True,
                    ),
                ),
                (
                    "parent_model",
                    models.CharField(editable=False, max_length=100, null=True),
                ),
                (
                    "priority",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("high", "High"),
                            ("medium", "Medium"),
                            ("low", "Low"),
                        ],
                        help_text="Leave blank to use default for this action type.",
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("New", "New"),
                            ("open", "Open"),
                            ("closed", "Closed"),
                            ("cancelled", "Cancelled"),
                        ],
                        default="New",
                        max_length=25,
                    ),
                ),
                (
                    "instructions",
                    models.TextField(
                        blank=True, help_text="populated by action class", null=True
                    ),
                ),
                ("auto_created", models.BooleanField(default=False)),
                (
                    "auto_created_comment",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        primary_key=True, serialize=False
                    ),
                ),
                (
                    "action_type",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionType",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "parent_action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.HistoricalActionItem",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical ",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
        ),
        migrations.CreateModel(
            name="HistoricalActionItemUpdate",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "report_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                ("comment", models.TextField(blank=True, max_length=250, null=True)),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        primary_key=True, serialize=False
                    ),
                ),
                (
                    "action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical ",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
        ),
        migrations.AddField(
            model_name="actionitem",
            name="action_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="action_type",
                to="edc_action_item.ActionType",
                verbose_name="Action",
            ),
        ),
        migrations.AddField(
            model_name="actionitem",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="actionitem",
            unique_together={("subject_identifier", "action_type", "reference_identifier")},
        ),
    ]



from django.apps import apps as django_apps
from django.db import models
from edc_constants.constants import HIGH_PRIORITY
from edc_model.models import BaseUuidModel

from ..choices import PRIORITY


class ActionTypeError(Exception):
    pass


class ActionTypeManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, name):
        return self.get(name=name)


class ActionType(BaseUuidModel):

    name = models.CharField(max_length=50, unique=True)

    display_name = models.CharField(max_length=100)

    reference_model = models.CharField(
        max_length=100, null=True, blank=True, help_text="reference model"
    )

    related_reference_model = models.CharField(
        max_length=100, null=True, blank=True, help_text="origin reference model"
    )

    priority = models.CharField(max_length=25, choices=PRIORITY, default=HIGH_PRIORITY)

    show_on_dashboard = models.BooleanField(default=False)

    show_link_to_changelist = models.BooleanField(default=False)

    create_by_action = models.BooleanField(
        default=True, help_text="This action may be created by another action"
    )

    create_by_user = models.BooleanField(
        default=True, help_text="This action may be created by the user"
    )

    instructions = models.TextField(max_length=250, null=True, blank=True)

    objects = ActionTypeManager()

    def __str__(self):
        return self.display_name

    def natural_key(self):
        return (self.name,)  # noqa

    @property
    def reference_model_cls(self):
        model_cls = None
        if self.reference_model:
            try:
                model_cls = django_apps.get_model(self.reference_model)
            except (LookupError, ValueError) as e:
                raise ActionTypeError(f"{e}. Got reference_model=`{self.reference_model}`")
        return model_cls

    def save(self, *args, **kwargs):
        self.display_name = self.display_name or self.name
        if self.reference_model and self.reference_model != "edc_action_item.reference":
            try:
                if not self.reference_model_cls.action_name:
                    raise ActionTypeError(
                        f"Model missing an action class. See {repr(self.reference_model_cls)}"
                    )
            except AttributeError as e:
                if "action_name" in str(e):
                    raise ActionTypeError(
                        f"Model not configured for Actions. Are you using the model mixin? "
                        f"See {repr(self.reference_model_cls)}. Got {e}"
                    )
                else:
                    raise
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        ordering = ["name"]
        indexes = [models.Index(fields=["id", "name"])]
        default_permissions = ("add", "change", "delete", "view", "export", "import")



# Generated by Django 3.0.6 on 2020-05-28 02:20

from django.db import migrations

import edc_action_item.models.action_type


class Migration(migrations.Migration):

    dependencies = [
        ("edc_action_item", "0024_auto_20191024_1000"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="actiontype",
            managers=[
                ("objects", edc_action_item.models.action_type.ActionTypeManager()),
            ],
        ),
    ]



import sys

from django.apps import AppConfig as DjangoApponfig
from django.core.management.color import color_style
from django.db.models.signals import post_migrate

from .site_action_items import site_action_items

style = color_style()


def update_action_types(sender=None, verbose=None, **kwargs):  # noqa

    sys.stdout.write(style.MIGRATE_HEADING("Updating action types:\n"))
    site_action_items.create_or_update_action_types()
    sys.stdout.write("Done.\n")
    sys.stdout.flush()


class AppConfig(DjangoApponfig):
    name = "edc_action_item"
    verbose_name = "Action Items"
    has_exportable_data = True
    include_in_administration_section = True

    def ready(self):
        post_migrate.connect(update_action_types, sender=self)
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        site_action_items.autodiscover()
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")



from datetime import datetime
from typing import Any, Optional, Type

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.deletion import PROTECT
from edc_constants.constants import NEW
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_notification.model_mixins import NotificationModelMixin
from edc_sites.models import CurrentSiteManager as BaseCurrentSiteManager
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow

from edc_action_item.stubs import ActionStub

from ..choices import ACTION_STATUS, PRIORITY
from ..identifiers import ActionIdentifier
from ..site_action_items import site_action_items
from .action_type import ActionType


class ActionItemUpdatesRequireFollowup(Exception):
    pass


class SubjectDoesNotExist(Exception):
    pass


class CurrentSiteManager(BaseCurrentSiteManager):

    use_in_migrations = True

    def get_by_natural_key(self, action_identifier):
        return self.get(action_identifier=action_identifier)


class ActionItemManager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, action_identifier):
        return self.get(action_identifier=action_identifier)


class ActionItem(
    NonUniqueSubjectIdentifierFieldMixin,
    SiteModelMixin,
    NotificationModelMixin,
    BaseUuidModel,
):

    subject_identifier_model = "edc_registration.registeredsubject"

    action_identifier = models.CharField(max_length=50, unique=True)

    report_datetime = models.DateTimeField(default=get_utcnow)

    action_type = models.ForeignKey(
        ActionType, on_delete=PROTECT, related_name="action_type", verbose_name="Action"
    )

    reference_model = models.CharField(
        max_length=50,
        null=True,
        help_text="model name in label_lower format",
    )

    linked_to_reference = models.BooleanField(
        default=False,
        editable=False,
        help_text=(
            "True if this action is linked to it's reference_model."
            "Initially False if this action is created before reference_model."
            "Always True when reference_model creates the action."
            'Set to True when reference_model is created and "links" to this action.'
            "(Note: reference_model looks for actions where "
            "linked_to_reference is False before attempting to "
            "create a new ActionItem)."
        ),
    )

    related_reference_model = models.CharField(max_length=100, null=True, editable=False)

    related_action_identifier = models.CharField(
        max_length=50,
        null=True,
        blank=True,
        help_text=(
            "May be left blank. e.g. action identifier from "
            "source model that opened the item."
        ),
    )

    parent_action_identifier = models.CharField(
        max_length=50,
        null=True,
        blank=True,
        help_text=(
            "May be left blank. e.g. action identifier from "
            "reference model that opened the item (parent)."
        ),
    )

    priority = models.CharField(
        max_length=25,
        choices=PRIORITY,
        null=True,
        blank=True,
        help_text="Leave blank to use default for this action type.",
    )

    parent_action_item = models.ForeignKey(
        "self",
        on_delete=PROTECT,
        related_name="+",
        null=True,
        blank=True,
        editable=False,
    )

    related_action_item = models.ForeignKey(
        "self",
        on_delete=PROTECT,
        related_name="+",
        null=True,
        blank=True,
        editable=False,
    )

    status = models.CharField(max_length=25, default=NEW, choices=ACTION_STATUS)

    instructions = models.TextField(
        null=True, blank=True, help_text="populated by action class"
    )

    auto_created = models.BooleanField(default=False)

    auto_created_comment = models.CharField(max_length=25, null=True, blank=True)

    on_site = CurrentSiteManager()

    objects = ActionItemManager()

    history = HistoricalRecords()

    def __str__(self):
        return (
            f"{self.action_type.display_name} {self.action_identifier[-9:]} "
            f"for {self.subject_identifier} ({self.get_status_display()})"
        )

    def save(self, *args, **kwargs):
        """See also signals and action_cls."""
        if not self.id:
            # a new persisted action item always has
            # a unique action identifier
            self.action_identifier = ActionIdentifier().identifier
            # subject_identifier
            subject_identifier_model_cls = django_apps.get_model(self.subject_identifier_model)
            try:
                subject_identifier_model_cls.objects.get(
                    subject_identifier=self.subject_identifier
                )
            except ObjectDoesNotExist:
                raise SubjectDoesNotExist(
                    f"Attempt to create {self.__class__.__name__} failed. "
                    f"Invalid subject identifier. Subject does not exist "
                    f"in '{self.subject_identifier_model}'. "
                    f"Got '{self.subject_identifier}'."
                )
            self.priority = self.priority or self.action_type.priority
            self.reference_model = self.action_type.reference_model
            self.related_reference_model = self.action_type.related_reference_model
            self.instructions = self.action_type.instructions
        super().save(*args, **kwargs)

    def natural_key(self) -> tuple:
        return (self.action_identifier,)  # noqa

    @property
    def last_updated(self) -> Optional[datetime]:
        return None if self.status == NEW else self.modified

    @property
    def user_last_updated(self) -> Optional[str]:
        return None if self.status == NEW else self.user_modified or self.user_created

    @property
    def action_cls(self) -> Type[ActionStub]:
        """Returns the action_cls."""
        return site_action_items.get(self.action_type.name)

    @property
    def action(self) -> ActionStub:
        """Returns the instantiated action_cls."""
        return self.action_cls(
            subject_identifier=self.subject_identifier,
            action_identifier=self.action_identifier,
            readonly=True,
        )

    @property
    def reference_model_cls(self) -> Type[BaseUuidModel]:
        return django_apps.get_model(self.reference_model)

    @property
    def reference_obj(self) -> Any:
        return self.reference_model_cls.objects.get(action_identifier=self.action_identifier)

    @property
    def parent_reference_obj(self: Any) -> Any:
        if not self.parent_action_item:
            raise ObjectDoesNotExist(
                f"Parent ActionItem does not exist for {self.action_identifier}."
            )
        return self.parent_action_item.reference_obj

    @property
    def related_reference_obj(self) -> BaseUuidModel:
        if not self.related_action_item:
            raise ObjectDoesNotExist(
                f"Related ActionItem does not exist for {self.action_identifier}."
            )
        return self.related_action_item.reference_obj

    @property
    def identifier(self) -> str:
        """Returns a shortened action identifier."""
        return self.action_identifier[-9:]

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Action Item"
        verbose_name_plural = "Action Items"
        indexes = [
            models.Index(
                fields=[
                    "id",
                    "action_identifier",
                    "action_type",
                    "status",
                    "report_datetime",
                ]
            )
        ]



# Generated by Django 2.2.5 on 2019-10-07 03:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_action_item", "0022_auto_20190628_2136")]

    operations = [
        migrations.AlterField(
            model_name="actionitem",
            name="subject_identifier",
            field=models.CharField(max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="subject_identifier",
            field=models.CharField(max_length=50),
        ),
    ]
